from .email import is_valid_email_address
from .results import Success, Failure
from inspect import signature


class Parameters(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self

    def with_attribute(self, parameter_name, value):
        return Parameters({**{parameter_name: value}, **self})


class ValidationResults:
    def __init__(self):
        self.valid_parameters = Parameters()

    def add(self, parameter_name, value):
        self.valid_parameters = self.valid_parameters.with_attribute(parameter_name, value)

    def map(self, fn):
        return fn(self.valid_parameters)


class Validator:
    def __init__(self, parameter_name, value):
        self.parameter_name = parameter_name
        self.value = value


class PresenceOf(Validator):
    def validate(self, results):
        if self.value is None:
            return Failure(message="{} is missing".format(self.parameter_name))
        results.add(self.parameter_name, self.value)
        return Success()


class OptionalityOf(Validator):
    def validate(self, results):
        results.add(self.parameter_name, self.value)
        return Success()


class MaxLengthOf(Validator):
    def __init__(self, parameter, value, max):
        super().__init__(parameter, value)
        self.max = max

    def validate(self, results):
        if len(self.value) > self.max: 
            return Failure(message="{} is too long".format(self.parameter_name))
        results.add(self.parameter_name, self.value)
        return Success()


class IsBetween(Validator):
    def __init__(self, parameter, value, lower_bound, upper_bound):
        super().__init__(parameter, value)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound

    def validate(self, results):
        return (int_from_string(self.value, self.parameter_name)
                .map(self._is_between)
                .do(lambda result: results.add(self.parameter_name, result.value)))

    def _is_between(self, result):
        if self.lower_bound <= result.value <= self.upper_bound:
            return Success(value=result.value) 
        return Failure(message='{} should be between {} and {}'.format(self.parameter_name, self.lower_bound, self.upper_bound))


class ConversionOf(Validator):
    def __init__(self, parameter, value, type_to_convert_to):
        super().__init__(parameter, value)
        self.type_to_convert_to = type_to_convert_to
    
    def validate(self, results):
        fn = self.type_to_convert_to.parse_from
        if 'success_attribute' in signature(fn).parameters:
            result = self.type_to_convert_to.parse_from(self.value, self.parameter_name)
        else:
            result = self.type_to_convert_to.parse_from(self.value)
        return result.do(lambda result: results.add(self.parameter_name, getattr(result, self.parameter_name)))


class ValidEmailAddress(Validator):
    def validate(self, results):
        stripped = self.value.strip()
        if not is_valid_email_address(stripped):
            return Failure(message="{} does not contain a valid email address".format(self.parameter_name))
        results.add(self.parameter_name, stripped)
        return Success()


class TryTo(Validator):
    def __init__(self, parameter_name, value, success_attr):
        super().__init__(parameter_name, value)
        self.success_attr = success_attr

    def validate(self, results):
        if self.value.is_failure():
            return self.value
        results.add(self.parameter_name, self.value.body[self.success_attr])
        return Success()


def presence_of(parameter, value):
    return PresenceOf(parameter, value)


def optionality_of(parameter, value):
    return OptionalityOf(parameter, value)


def max_length_of(parameter, value, max):
    return MaxLengthOf(parameter, value, max)


def is_between(parameter, value, lower_bound, upper_bound):
    return IsBetween(parameter, value, lower_bound, upper_bound)


def conversion_of(parameter, value, type_to_convert_to):
    return ConversionOf(parameter, value, type_to_convert_to)


def email_validity_of(parameter, value):
    return ValidEmailAddress(parameter, value)


def an_attempt_to(value, parameter, success_attr):
    return TryTo(parameter, value, success_attr=success_attr)


def validate(*validations):
    validation_results = ValidationResults()
    for validation in validations:
        last_result = validation.validate(validation_results)
        if last_result.is_failure():
            return last_result
    return validation_results


def int_from_string(string_value, name):
    if string_value is None: return Failure(message="{} is missing".format(name))
    try: 
        return Success(value=int(string_value))
    except ValueError:
        return Failure(message='{} is not an integer value'.format(name))



import uuid
from uuid import UUID as UUID_IMPL
from dataclasses import dataclass, field


@dataclass(frozen=True)
class ID:
    _uuid: UUID_IMPL

    @staticmethod
    def from_string(potential_uuid_value):
        if not potential_uuid_value:
            return InvalidID()
        try:
            return ID(UUID_IMPL(potential_uuid_value))
        except ValueError as e:
            return InvalidID()

    @staticmethod
    def invalid():
        return InvalidID()

    @property
    def valid(self):
        return True

    def __str__(self):
        return str(self._uuid)


class IDGenerator(object):
    def generate_id(self):
        return ID(uuid.uuid4())


class FixedIDGeneratorGenerating(IDGenerator):
    def __init__(self, *ids):
        for id in ids:
            assert id.__class__ == ID
        self.ids = ids
        self.index = -1

    def generate_id(self):
        self.index = (self.index + 1) % len(self.ids)
        return self.ids[self.index]


class InvalidValidGenerator:
    counter = 0

    @staticmethod
    def next_invalid_value():
        print('generating')
        return 'invalid_id_{}'.format(InvalidValidGenerator.next())

    @staticmethod
    def next():
        InvalidValidGenerator.counter += 1
        return InvalidValidGenerator.counter


@dataclass(frozen=True)
class InvalidID:
    value: str = field(default_factory=InvalidValidGenerator.next_invalid_value)

    @property
    def valid(self):
        return False

    def __str__(self):
        return self.value



from dataclasses import dataclass


@dataclass(init=False)
class Result:
    body: any

    def __init__(self, **kwargs):
        self.body = kwargs

    def __getattr__(self, name):
        return self.body.get(name)
    
    def with_attributes(self, **kwargs):
        return self.__class__(**{**self.body, **kwargs})


class Success(Result):
    def is_success(self):
        return True

    def is_failure(self):
        return False

    def map(self, f):
        return f(self)

    def do(self, f):
        f(self)
        return self

    def prefix_error(self, prefix):
        return self

    def or_fail_with(self, **kwargs):
        return self


class Failure(Result):
    def is_success(self):
        return False

    def is_failure(self):
        return True

    def map(self, f):
        return self

    def prefix_error(self, prefix):
        return Failure(message=prefix + self.message)

    def do(self, f):
        return self

    def or_fail_with(self, **kwargs):
        return Failure(**dict(reason=self), **kwargs)


class PartialSuccess(Result):
    def is_success(self):
        return False

    def is_failure(self):
        return True

    def map(self, f):
        return PartialSuccess(**{**self.body, **f(self).body})

    def do(self, f):
        f(self)
        return self

    def or_fail_with(self, **kwargs):
        return self



from datetime import date
from .results import Success, Failure


class StringToDateParser():
    def parse_from(self, date_string, success_attribute='date'):
        if not date_string:
            return Failure(message='date is missing')
        try:
            return Success(**{success_attribute: date.fromisoformat(date_string)})
        except ValueError as e:
            return Failure(message=str(e))


class StringToIntParser:
    def parse_from(self, int_string, success_attribute='int_val'):
        try:
            return Success(**{success_attribute: int(int_string)})
        except ValueError as e:
            return Failure(message="'{}' is not a valid integer".format(int_string))


date_from_iso = StringToDateParser()


int_from_string = StringToIntParser()



import re


valid_email_addresses = re.compile(r'^[^.]([\w!#$%&\\\'*+\-/=?^_`.{|}~]*[^.])?@([\w-]+\.)+[\w-]+$')
double_dots = re.compile(r'\.\.')


def is_valid_email_address(value):
    return valid_email_addresses.match(value) is not None and double_dots.search(value) is None



from setuptools import setup, find_packages
import sys
sys.path.append('src')
from quiltz.domain.version import version

with open("README.md", "r") as fh:
  long_description = fh.read()

setup(
  name='quiltz-domain',  
  version=version,
  author="Rob Westgeest",
  author_email="rob@qwan.eu",
  description="A domain utility module for python",
  long_description=long_description,
  long_description_content_type="text/markdown",
  url="https://github.com/qwaneu/quiltz-domain",
  packages=find_packages(where='src'),
  package_dir={'':'src'},
  classifiers=[
      "Programming Language :: Python :: 3",
      "License :: OSI Approved :: MIT License",
      "Operating System :: OS Independent",
  ],
)



from quiltz.domain.id import ID 
import uuid


def aValidUUID(simpleIdValue):
    return uuid.UUID("{:>32}".format(simpleIdValue).replace(' ', '1'))


def aValidID(simpleIdValue):
    return ID(aValidUUID(simpleIdValue))
